const { getFeatureNotSupportedError } = require('../../../util/errors')
const { getOnCond } = require('../../../../common/utils/generateOnCond')

const isNavigation = pathSegments => {
  return pathSegments.length > 1 && pathSegments[1].getKind().startsWith('NAVIGATION')
}

const isViewWithParams = target => {
  return target.params && Object.keys(target.params).length > 0
}

const _entityNameFromSegment = segment => {
  if (segment.getKind() === 'SINGLETON') {
    return segment.getSingleton().getEntityType().getFullQualifiedName().toString()
  }

  if (segment.getKind() === 'ENTITY') {
    return segment.getEntitySet().getEntityType().getFullQualifiedName().toString()
  }

  return segment.getNavigationProperty().getEntityType().getFullQualifiedName().toString()
}

const _keysFromSegment = segment => {
  if (segment.getKeyPredicates().length > 0) {
    const keys = {}

    for (const keyPredicate of segment.getKeyPredicates()) {
      const key = keyPredicate.getEdmRef().getName().replace(/\//g, '.')
      keys[key] = keyPredicate.getText()
    }

    return keys
  }
}

const _addKeysToWhereIfNeeded = (cqn, keys, tableAlias) => {
  if (keys) {
    for (const key in keys) {
      cqn.where([{ ref: [`${tableAlias}`, `${key}`] }, '=', { val: keys[key] }])
    }
  }
}

const _addOnCondToWhere = (cqn, entity, tableAlias, identifier, csn) => {
  const onConditionOptions = {
    associationNames: entity.current,
    csn: csn,
    aliases: {
      select: tableAlias,
      join: identifier
    }
  }

  const onCond = getOnCond(csn.definitions[entity.previous].elements[entity.current], onConditionOptions)
  cqn.where(onCond)
}

const enhanceCqnWithSubSelects = (cqn, pathSegments, csn, SELECT) => {
  let previousCqn, previousEntityName

  // if .../property or .../$count requested, must be ignored when building query
  const segments = pathSegments.filter(s => s.getProperty() === null && s.getKind() !== 'COUNT')

  for (let i = 0; i < segments.length; i++) {
    const isLastElement = i === segments.length - 1
    const tableAlias = `T${i}`
    const entityName = _entityNameFromSegment(segments[i])
    const keys = _keysFromSegment(segments[i])
    let currentCqn

    if (isLastElement) {
      cqn._from(entityName, tableAlias)
      _addKeysToWhereIfNeeded(cqn, keys, tableAlias)
    } else {
      currentCqn = SELECT.from(`${entityName} as ${tableAlias}`, [1])
      _addKeysToWhereIfNeeded(currentCqn, keys, tableAlias)
    }

    if (previousCqn) {
      _addOnCondToWhere(
        previousCqn,
        { current: segments[i].getNavigationProperty().getName(), previous: previousEntityName },
        tableAlias,
        `T${i - 1}`,
        csn
      )

      if (isLastElement) {
        cqn.where(['exists', previousCqn])
      } else {
        currentCqn.where(['exists', previousCqn])
      }
    }

    previousCqn = currentCqn
    previousEntityName = entityName
  }
}

const validationQuery = (pathSegments, csn, SELECT) => {
  const cqn = SELECT.from('placeholder')
  enhanceCqnWithSubSelects(cqn, pathSegments.slice(0, pathSegments.length - 1), csn, SELECT)
  return cqn
}

const isPathSupported = (supported, pathSegments) => {
  for (const segment of pathSegments) {
    if (!supported[segment.getKind()]) {
      throw getFeatureNotSupportedError(`Request parameter "${segment.getKind()}"`)
    }
  }
}

module.exports = {
  enhanceCqnWithSubSelects,
  isNavigation,
  isViewWithParams,
  isPathSupported,
  validationQuery
}
