'use strict'

const commons = require('../../odata-commons')
const ResourceKind = commons.uri.UriResource.ResourceKind
const HttpMethods = commons.http.HttpMethod.Methods
const NotImplementedError = commons.errors.NotImplementedError
const MethodNotAllowedError = require('../errors/MethodNotAllowedError')
const Components = require('./ComponentManager').Components

/**
 * Validate if the requested method is included in an array of one ore more allowed methods.
 *
 * @param {string} requestMethod The request method.
 * @param {...string} allowedMethods One or more allowed methods.
 * @private
 */
function validateHttpMethod (requestMethod, ...allowedMethods) {
  for (const allowedMethod of allowedMethods) {
    if (requestMethod === allowedMethod) {
      return
    }
  }
  throw new MethodNotAllowedError()
}

/**
 * The main odata dispatcher for resource dispatching. The dispatcher resolves the appropriate
 * handler for the request and is responsible for its execution.
 */
class Dispatcher {
  /**
   * Sets the logger.
   * @param {LoggerFacade} logger the logger
   * @returns {Dispatcher} this instance
   */
  setLogger (logger) {
    this._logger = logger
    return this
  }

  /**
   * Sets the performance monitor.
   * @param {PerformanceMonitor} performanceMonitor the performance monitor
   * @returns {Dispatcher} this instance
   */
  setPerformanceMonitor (performanceMonitor) {
    this._performanceMonitor = performanceMonitor
    return this
  }

  /**
   * Dispatch the request to the appropriate handler.
   *
   * @param {OdataRequest} request the current OData request
   * @param {OdataResponse} response the current OData response
   * @param {ComponentManager} componentManager The component manager with registered components.
   * @returns {Promise} Resolve with execution result data or rejects with an error
   */
  dispatch (request, response, componentManager) {
    const lastKind = request
      .getUriInfo()
      .getLastSegment()
      .getKind()

    this._logger.debug('Found kind of last uriInfo segment:', lastKind)

    switch (lastKind) {
      case ResourceKind.METADATA:
        validateHttpMethod(request.getMethod(), HttpMethods.GET)
        return this._handle(request, response, componentManager, Components.METADATA_HANDLER)

      case ResourceKind.SERVICE:
        validateHttpMethod(request.getMethod(), HttpMethods.GET, HttpMethods.HEAD)
        return this._handle(request, response, componentManager, Components.SERVICE_HANDLER)

      case ResourceKind.BATCH:
        validateHttpMethod(request.getMethod(), HttpMethods.POST)
        return this._handle(request, response, componentManager, Components.BATCH_EXECUTE_HANDLER)

      case ResourceKind.BOUND_ACTION:
      case ResourceKind.ACTION_IMPORT:
        validateHttpMethod(request.getMethod(), HttpMethods.POST)
        return this._handle(request, response, componentManager, Components.ACTION_EXECUTE_HANDLER)

      default:
        break
    }

    switch (request.getMethod()) {
      case HttpMethods.GET:
        return this._handle(request, response, componentManager, Components.DATA_READ_HANDLER)

      case HttpMethods.DELETE:
        return this._handle(request, response, componentManager, Components.DATA_DELETE_HANDLER)

      case HttpMethods.POST:
        return this._handle(request, response, componentManager, Components.DATA_CREATE_HANDLER)

      case HttpMethods.PATCH:
      case HttpMethods.PUT:
        return this._handle(request, response, componentManager, Components.DATA_UPDATE_HANDLER)

      case HttpMethods.HEAD:
        throw new NotImplementedError()

      default:
        throw new MethodNotAllowedError()
    }
  }

  /**
   * Handle a request with the selected handler.
   *
   * @param {OdataRequest} request the current OData request
   * @param {OdataResponse} response the current OData response
   * @param {ComponentManager} componentManager The component manager with registered components.
   * @param {string} handlerName The name of the handler.
   * @returns {Promise} Resolve with execution result data or rejects with an error
   * @private
   */
  _handle (request, response, componentManager, handlerName) {
    const handler = componentManager.getComponent(handlerName)
    this._logger.debug('Handler found for name', handlerName, handler)

    if (!handler) {
      throw new NotImplementedError(`Not implemented: Handler '${handlerName}' could not be found`)
    }

    const performanceMonitor = this._performanceMonitor
      .getChild('Handler dispatching')
      .createChild('Process handler ' + handlerName)
      .start()

    return new Promise((resolve, reject) => {
      handler(request, response, (err, resultData, resultOptions) => {
        this._logger.debug('Options result of handler operation:', resultOptions)

        return err ? reject(err) : resolve({ data: resultData, options: resultOptions })
      })
    }).then(result => {
      performanceMonitor.stop()
      return result
    })
  }
}

module.exports = Dispatcher
