'use strict'

const EventListenerCommand = require('./EventListenerCommand')
const OdataVersionValidationCommand = require('./OdataVersionValidationCommand')
const UriParserCommand = require('./UriParserCommand')
const ParsePreferHeaderCommand = require('./ParsePreferHeaderCommand')
const OperationValidationCommand = require('./OperationValidationCommand')
const QueryOptionsValidationCommand = require('./QueryOptionsValidationCommand')
const QueryOptionsParserCommand = require('./QueryOptionsParserCommand')
const ConditionalRequestPreValidationCommand = require('./ConditionalRequestPreValidationCommand')
const ContentNegotiatorCommand = require('./ContentNegotiatorCommand')
const LocaleNegotiatorCommand = require('./LocaleNegotiatorCommand')
const PresetResponseHeadersCommand = require('./PresetResponseHeadersCommand')
const DeserializingCommand = require('./DeserializingCommand')
const RequestContentValidationCommand = require('./RequestContentValidationCommand')
const DispatcherCommand = require('./DispatcherCommand')
const ConditionalRequestControlCommand = require('./ConditionalRequestControlCommand')
const SetResponseHeadersCommand = require('./SetResponseHeadersCommand')
const SetStatuscodeCommand = require('./SetStatuscodeCommand')
const SerializingCommand = require('./SerializingCommand')
const SendResponseCommand = require('./SendResponseCommand')
const ErrorSerializingCommand = require('./ErrorSerializingCommand')
const ErrorContentNegotiatorCommand = require('./ErrorContentNegotiatorCommand')
const DebugDeserializingCommand = require('./DebugDeserializingCommand')
const DebugContentNegotiatorCommand = require('./DebugContentNegotiatorCommand')
const DebugSerializingCommand = require('./DebugSerializingCommand')

const RequestContentNegotiator = require('../format/RequestContentNegotiator')
const HttpStatusCode = require('../http/HttpStatusCode')
const Components = require('../core/ComponentManager').Components

/**
 * The CommandFactory creates chains of commands which can be used for execution by a corresponding executor.
 */
class CommandFactory {
  /**
   * Creates a chain of commands which can be used to execute a default single request.
   * If the URL indicates that the request should be in debug mode,
   * additional debug commands to handle the debug mode will be added to the command chain.
   *
   * @param {Object} options An options object with configuration properties
   * @param {ComponentManager} options.componentManager The current component manager instance
   * @param {FormatManager} options.formatManager The current format manager instance
   * @param {Context} options.context The current odata context instance
   * @param {boolean} options.isDebug If true request will be processed in debug mode
   * @returns {Array.<Array>} an array of commands with their descriptions
   */
  static createForSingleRequest (options) {
    const context = options.context
    const request = context.getRequest()
    const response = context.getResponse()
    const service = context.getService()
    const logger = context.getLogger()

    const uriParser = options.componentManager
      .getComponent(Components.URI_PARSER)
      .setPerformanceMonitor(context.getPerformanceMonitor())
    const contentNegotiator = options.componentManager.getComponent(Components.CONTENT_NEGOTIATOR).setLogger(logger)
    const localeNegotiator = options.componentManager.getComponent(Components.LOCALE_NEGOTIATOR)
    const requestValidator = options.componentManager.getComponent(Components.REQUEST_VALIDATOR).setLogger(logger)
    const dispatcher = options.componentManager
      .getComponent(Components.DISPATCHER)
      .setLogger(logger)
      .setPerformanceMonitor(context.getPerformanceMonitor())

    let result = [
      [
        new EventListenerCommand(context, options.componentManager.getComponent('request'), 'request'),
        'Eventlistener request'
      ],
      [
        new OdataVersionValidationCommand(request, requestValidator, null, logger),
        'Initial request version-header validation'
      ],
      [new UriParserCommand(uriParser, request, logger), 'Uri parsing'],
      [new ParsePreferHeaderCommand(request, requestValidator, logger), 'Prefer parsing'],
      [new OperationValidationCommand(request, requestValidator, logger), 'Operation validation'],
      [new QueryOptionsValidationCommand(requestValidator, request, response, logger), 'Query options validation'],
      [new QueryOptionsParserCommand(uriParser, request, logger), 'Query options parsing'],
      [new ConditionalRequestPreValidationCommand(request, logger), 'Conditional request validation'],
      [
        new ContentNegotiatorCommand(request, response, options.formatManager, contentNegotiator, logger),
        'Content negotiation'
      ],
      [new LocaleNegotiatorCommand(request, response, localeNegotiator, logger), 'Locale negotiation'],
      [
        new PresetResponseHeadersCommand(request, response, service.getOdataVersion(), logger),
        'Preset response headers'
      ],
      [
        new DeserializingCommand(request, options.bodyParserManager, new RequestContentNegotiator(), logger),
        'Deserialize request body'
      ],
      [
        new RequestContentValidationCommand(request, requestValidator, service.getOdataVersion(), logger),
        'Request content validation'
      ],
      [
        new DispatcherCommand(
          request,
          response,
          service._getMetadataCache(),
          options.componentManager,
          dispatcher,
          logger
        ),
        'Handler dispatching'
      ],
      [
        new ConditionalRequestControlCommand(request, response, service.getFormatManager(), logger),
        'Conditional request control'
      ],
      [new SetResponseHeadersCommand(request, response, service.getFormatManager(), logger), 'Set response headers'],
      [new SetStatuscodeCommand(request, response, HttpStatusCode.resolveSuccessStatusCode, logger), 'Set status code'],
      [new SerializingCommand(context), 'Serializing'],
      [new SendResponseCommand(response, logger), 'Send response']
    ]
    if (options.isDebug) CommandFactory._addDebugCommands(result, options)
    return result
  }

  /**
   * Creates a chain of commands which can be used to execute an error event.
   * If the URL indicates that the request should be in debug mode,
   * additional debug commands to handle the debug mode will be added to the command chain.
   *
   * @param {Object} options An options object with configuration properties
   * @param {Object} options.componentManager The current component manager instance
   * @param {Object} options.formatManager The current format manager instance
   * @param {Object} options.context The current odata context instance
   * @param {boolean} options.isDebug If true request will be processed in debug mode
   * @returns {Array.<Array>} an array of commands with their descriptions
   */
  static createForSingleRequestError (options) {
    const context = options.context
    const request = context.getRequest()
    const response = context.getResponse()
    const logger = context.getLogger()

    const contentNegotiator = options.componentManager.getComponent(Components.CONTENT_NEGOTIATOR).setLogger(logger)

    let result = [
      [
        new EventListenerCommand(context, options.componentManager.getComponent('error'), 'error'),
        'Eventlistener error'
      ],
      [
        new ErrorContentNegotiatorCommand(request, response, options.formatManager, contentNegotiator, logger),
        'Error content negotiation'
      ],
      [
        new PresetResponseHeadersCommand(request, response, context.getService().getOdataVersion(), logger),
        'Set error response headers'
      ],
      [
        new SetStatuscodeCommand(request, response, HttpStatusCode.resolveErrorStatusCode, logger),
        'Set error status code'
      ],
      [new ErrorSerializingCommand(response, logger), 'Serializing error'],
      [new SendResponseCommand(response, logger), 'Send error response']
    ]
    if (options.isDebug) CommandFactory._addDebugCommands(result, options)
    return result
  }

  /**
   * Adds debug negotiation and debug serializing commands to the provided chain of commands.
   * @param {Array.<Array>} commands Array of commands with their descriptions to add the debug commands to
   * @param {Object} options An options object with configuration properties
   * @param {Object} options.formatManager The current format manager instance
   * @param {Object} options.context The current odata context instance
   */
  static _addDebugCommands (commands, options) {
    const context = options.context
    const request = context.getRequest()
    const response = context.getResponse()
    const logger = context.getLogger()

    const contentNegotiator = options.componentManager.getComponent(Components.CONTENT_NEGOTIATOR).setLogger(logger)

    // For debug output, we have to stop the runtime measurement prematurely;
    // otherwise we won't get the total runtime in the output (which is written in a command, too).
    // We do that in the first command of the debug command chain.
    commands.push([new DebugDeserializingCommand(request, logger), 'Debug deserialize request body'])

    commands.push([
      new DebugContentNegotiatorCommand(request, response, options.formatManager, contentNegotiator, logger),
      'Debug content negotiation'
    ])
    commands.push([
      new SetStatuscodeCommand(request, response, HttpStatusCode.resolveDebugStatusCode, logger),
      'Debug set status code'
    ])
    commands.push([new DebugSerializingCommand(context), ''])
    commands.push([
      new PresetResponseHeadersCommand(request, response, context.getService().getOdataVersion(), logger),
      ''
    ])
    commands.push([new SendResponseCommand(response, logger), ''])
  }
}

module.exports = CommandFactory
