'use strict'

const commons = require('../../odata-commons')
const Methods = commons.http.HttpMethod.Methods
const ResourceKinds = commons.uri.UriResource.ResourceKind
const Command = require('./Command')
const ConditionalRequestValidator = require('../validator/ConditionalRequestValidator')

/**
 * The `next` callback to be called upon finish execution.
 *
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Executes the validation of the request conditional request.
 */
class ConditionalRequestValidationCommand extends Command {
  /**
   * Creates an instance of ConditionalRequestValidationCommand.
   *
   * @param {OdataRequest} request the current OData request
   * @param {LoggerFacade} logger the logger
   */
  constructor (request, logger) {
    super()
    this._request = request
    this._logger = logger
  }

  execute (next) {
    const uriInfo = this._request.getUriInfo()
    const isConcurrentResource = this._request.getConcurrentResource() !== null

    this._logger.debug('Requested resource is concurrent:', isConcurrentResource)

    const ifNoneMatch = this._request.getHeader('if-none-match')
    const ifMatch = this._request.getHeader('if-match')
    const method = this._request.getMethod()
    const isValidOperation =
      method === Methods.GET ||
      method === Methods.PUT ||
      method === Methods.PATCH ||
      method === Methods.DELETE ||
      (method === Methods.POST && uriInfo.getLastSegment().getKind() === ResourceKinds.REF_COLLECTION) ||
      uriInfo.getLastSegment().getKind() === ResourceKinds.BOUND_ACTION

    if (isValidOperation && (ifMatch || ifNoneMatch || isConcurrentResource)) {
      this._logger.debug('Header If-Match:', "'" + ifMatch + "'", 'If-None-Match:', "'" + ifNoneMatch + "'")

      new ConditionalRequestValidator().preValidate(ifMatch, ifNoneMatch, method, isConcurrentResource)
    }

    next()
  }
}

module.exports = ConditionalRequestValidationCommand
