'use strict'

const Command = require('./Command')

/**
 * The `next` callback to be called upon finish execution.
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Negotiates the response content type for the provided request.
 * @extends Command
 */
class ContentNegotiatorCommand extends Command {
  /**
   * Creates an instance of ContentNegotiatorCommand.
   * @param {OdataRequest} request the current OData request
   * @param {OdataResponse} response the current OData response
   * @param {FormatManager} formatManager The current instance of format manager
   * @param {ResponseContentNegotiator} contentNegotiator The current instance of ResponseContentNegotiator
   * @param {LoggerFacade} logger the logger
   */
  constructor (request, response, formatManager, contentNegotiator, logger) {
    super()
    this._request = request
    this._response = response
    this._formatManager = formatManager
    this._negotiator = contentNegotiator
    this._logger = logger
  }

  /**
   * Executes the content negotiation. The content negotiation creates a `ResponseContract` object
   * as a result with all necessary content negotiation information. The contract object is
   * attached to the odata response instance.
   *
   * @param {Next} next The next callback to be called on finish
   */
  execute (next) {
    const contract = this._negotiator.negotiate(this._formatManager, this._request, this._request.getUriInfo())
    if (contract.getContentTypeInfo()) {
      this._logger.debug('Response contract content type:', contract.getContentTypeInfo().toString())
    }

    this._response.setContract(contract)

    next()
  }

  /**
   * Returns the current content negotiator instance.
   * @returns {ContentNegotiator} the current instance of the negotiator
   * @protected
   */
  getNegotiator () {
    return this._negotiator
  }

  /**
   * Returns the current instance of the format manager.
   * @returns {FormatManager} the current instance of the format manager
   * @protected
   */
  getFormatManager () {
    return this._formatManager
  }
}

module.exports = ContentNegotiatorCommand
