'use strict'

const RepresentationKinds = require('../../odata-commons').format.RepresentationKind.Kinds
const Command = require('./Command')
const DeserializationError = require('../errors/DeserializationError')

/**
 * The `next` callback to be called upon finish execution.
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Parses the provided request body.
 * @extends Command
 */
class DeserializingCommand extends Command {
  /**
   * Creates an instance of DeserializingCommand.
   * @param {OdataRequest} request the current OData request
   * @param {FormatManager} formatManager The current used format manager
   * @param {RequestContentNegotiator} negotiator The current used request payload negotiator
   * @param {LoggerFacade} logger the logger
   */
  constructor (request, formatManager, negotiator, logger) {
    super()
    this._request = request
    this._formatManager = formatManager
    this._negotiator = negotiator
    this._logger = logger
  }

  /**
   * Executes the request-body parsing. The content negotiation creates a `RequestContract` object
   * as a result with a payload deserializer facade inside. The contract object is
   * attached to the odata request instance. The deserializer facade is executed and the result body
   * is attached to the request. This command is executed only if the incoming request has a body
   * to parse as determined by the negotiation.
   *
   * @param {Next} next The next callback to be called on finish
   */
  execute (next) {
    const contract = this._negotiator.negotiate(this._formatManager, this._request)
    const representationKind = contract.getRepresentationKind()
    if (representationKind === RepresentationKinds.NO_CONTENT) {
      next()
    } else {
      const deserializerFacade = contract.getDeserializerFunction()
      if (deserializerFacade) {
        deserializerFacade(this._request, (err, body) => {
          this._request.setBody(body)
          next(err)
        })
      } else {
        next(
          new DeserializationError(
            "No payload deserializer available for resource kind '" +
              representationKind +
              "' and mime type '" +
              contract.getContentTypeInfo().getMimeType() +
              "'"
          )
        )
      }
    }
  }
}

module.exports = DeserializingCommand
