'use strict'

const Command = require('./Command')
const RepresentationKind = require('../../odata-commons').format.RepresentationKind.Kinds

/**
 * The `next` callback to be called upon finish execution.
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Calls the application in order to negotiate the Locale of the response.
 * @extends Command
 */
class LocaleNegotiatorCommand extends Command {
  /**
   * Creates an instance of LocaleNegotiatorCommand.
   * @param {OdataRequest} request the current OData request
   * @param {OdataResponse} response the current OData response
   * @param {Function} localeNegotiator locale negotiator
   * @param {LoggerFacade} logger the logger
   */
  constructor (request, response, localeNegotiator, logger) {
    super()
    this._request = request
    this._response = response
    this._localeNegotiator = localeNegotiator
    this._logger = logger
  }

  /**
   * Executes the locale negotiation.
   * @param {Next} next The next callback to be called on finish
   */
  execute (next) {
    const contract = this._response.getContract()
    if (contract.getRepresentationKind() === RepresentationKind.METADATA && this._localeNegotiator) {
      this._localeNegotiator(this._request, this._response, (error, localeInfo) => {
        if (error) {
          next(error)
        } else {
          if (localeInfo) {
            contract.setLocale(localeInfo.value)
            this._logger.debug('Response Contract Locale:', localeInfo.value)
          }
          next()
        }
      })
    } else {
      next()
    }
  }
}

module.exports = LocaleNegotiatorCommand
