'use strict'

const HttpMethods = require('../../odata-commons').http.HttpMethod.Methods
const Command = require('./Command')

/**
 * The `next` callback to be called upon finish execution.
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Executes the system query options validation.
 * @extends Command
 */
class QueryOptionsValidationCommand extends Command {
  /**
   * Creates an instance of QueryOptionsValidationCommand.
   * @param {RequestValidator} validator the request validator
   * @param {OdataRequest} request the OData request
   * @param {OdataResponse} response the OData response
   * @param {LoggerFacade} logger the logger
   */
  constructor (validator, request, response, logger) {
    super()
    this._validator = validator
    this._request = request
    this._response = response
    this._logger = logger
  }

  /**
   * Validates the query options.
   * @param {Next} next The next callback to be called on finish
   */
  execute (next) {
    const uriInfo = this._request.getUriInfo()
    const queryOptions = this._request.getQueryOptions()

    const method = this._request.getMethod()
    if (method === HttpMethods.GET) {
      this._validator.validateQueryOptions(queryOptions, uriInfo)
    } else {
      this._validator.validateQueryOptionsForNonGetHttpMethod(queryOptions, method, uriInfo)
    }

    this._validator.checkForForbiddenQueryOptions(queryOptions)

    try {
      this._validator.validateDebugOption(queryOptions)
      next()
    } catch (error) {
      // We disable debug mode because there is an error with debug option itself.
      // This must be done to avoid further buffering of the response.
      this._request.getContract().enableDebugMode(false)
      this._response.setBuffered(false)
      next(error)
    }
  }
}

module.exports = QueryOptionsValidationCommand
