'use strict'

const HttpMethods = require('../../odata-commons').http.HttpMethod.Methods
const Command = require('./Command')
const MethodNotAllowedError = require('../errors/MethodNotAllowedError')

/**
 * The `next` callback to be called upon finish execution.
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Executes the validation of the request content.
 * @extends Command
 */
class RequestContentValidationCommand extends Command {
  /**
   * Creates an instance of the RequestContentValidationCommand.
   * @param {OdataRequest} request the current OData request
   * @param {RequestValidator} requestValidator The current request validator
   * @param {?string} version the supported OData version
   * @param {LoggerFacade} logger the logger
   */
  constructor (request, requestValidator, version, logger) {
    super()
    this._request = request
    this._requestValidator = requestValidator
    this._version = version
    this._logger = logger
  }

  /**
   * Executes the validation.
   * @param {Next} next The next callback to be called on finish
   */
  execute (next) {
    const method = this._request.getMethod()
    const isDeepUpdate = this._request.getDeepInsertExpand().length && method !== HttpMethods.POST
    const allowedVersion = this._request.hasDelta() || isDeepUpdate ? '4.01' : this._version
    this._requestValidator.validateVersion(allowedVersion, this._request.getHeaders())

    if (this._request.hasDelta() && method !== HttpMethods.PATCH) {
      throw new MethodNotAllowedError('Requests with delta annotations must use HTTP PATCH')
    }

    next()
  }
}

module.exports = RequestContentValidationCommand
