'use strict'

const Command = require('./Command')
const InternalServerError = require('../errors/InternalServerError')

/**
 * The `next` callback to be called upon finish execution.
 * @callback Next
 * @param {?Error} error An error if there is one or null if not
 */

/**
 * Executes the setting of the statuscode in the odata response
 */
class SetStatuscodeCommand extends Command {
  /**
   * Creates an instance of SetStatuscodeCommand.
   * @param {OdataRequest} request the current OData request
   * @param {OdataResponse} response the current OData response
   * @param {Function} resolveFunction The function to resolve the statuscode with parameters context and error
   * @param {LoggerFacade} logger the logger
   */
  constructor (request, response, resolveFunction, logger) {
    super()
    this._request = request
    this._response = response
    this._resolveStatusCodeFunction = resolveFunction
    this._logger = logger
  }

  /**
   * @param {next} next The next callback to be called on finish
   * @param {?Error} error An error if there is one or null if not
   */
  execute (next, error) {
    const statusCode = this._resolveStatusCodeFunction(this._request, this._response, error)

    if (statusCode) {
      const options = error || this._request.getContract().isDebug() ? { overwrite: true } : undefined

      this._logger.debug('Set response status code:', statusCode, ', options:', options)
      this._response.setStatusCode(statusCode, options)

      next()
    } else {
      next(new InternalServerError('Invalid state: status code is not defined'))
    }
  }
}

module.exports = SetStatuscodeCommand
