'use strict'

const CRLF = '\r\n'

/**
 * Writes content in application/multipart format into an stream.
 * Used to write a OData batch response. Usage:
 *
 * E.g.:
 * '--boundary',                                            <-- not written by this class
 * 'Content-ID: 2',                                         <-- written by ApplicationHttpWriter->writeHeader
 * 'Content-Type: application/http',
 * '',
 * "PATCH /service/Customers('ALFKI') HTTP/1.1'",           <-- written by ApplicationHttpWriter->write(response)
 * 'Host: host',
 * 'Content-Type: application/json',
 * '',
 * '{}',
 * '--boundary--',                                          <-- not written by this class
 */
class ApplicationHttpWriter {
  /**
   * Constructor
   * @param {Stream} writableStream Destination stream
   */
  constructor (writableStream) {
    this._writableStream = writableStream

    this._appHeaderWritten = false
  }

  /**
   * Writes the content type into the destination stream if not already done.
   */
  writeContentType () {
    if (!this._appHeaderWritten) {
      this._appHeaderWritten = true
      this._writableStream.write('content-type: application/http' + CRLF, 'utf8')
      this._writableStream.write('content-transfer-encoding: binary' + CRLF, 'utf8')
    }
  }

  /**
   * Writes a http header (name, value) into the destination stream
   *
   * @param {string} name Header name
   * @param {string} value Header value
   */
  writeHeader (name, value) {
    this.writeContentType()
    this._writableStream.write(name + ': ' + value + CRLF, 'utf8')
  }

  /**
   * Write a response object into the destination stream
   * @param {PlainHttpResponse} response the response object
   */
  write (response) {
    this.writeContentType()
    this._writableStream.write(CRLF)
    response.writeTo(this._writableStream)
  }
}

module.exports = ApplicationHttpWriter
