'use strict'

const MultipartWriter = require('./MultipartWriter')
const ApplicationHttpWriter = require('./ApplicationHttpWriter')

/**
 * Serializes the list of batched responses into the outgoing response which is sent to the client.
 */
class BatchMultipartSerializer {
  /**
   * @param {BatchContext} batchContext the batch context
   */
  constructor (batchContext) {
    this._batchContext = batchContext
    this._outgoingResponse = this._batchContext.getResponse()
  }

  /**
   * Creates a 2 layered list to ease writing the batched responses into the outgoing response.
   * The first layer contains batched responses or and array of batched responses.
   * In the second case the array contains the responses of an atomicity group.
   *
   * @returns {Array.<OdataResponseInBatch|OdataResponseInBatch[]>} the responses
   * @private
   */
  _createSendList () {
    const responsesInBatch = this._batchContext.getResponses()
    let lastUsedAtomicityGroup = null
    let list = []
    let grp = null

    // Loop over requests to ensure that responses are serialized in request order.
    for (const request of this._batchContext.getRequestList()) {
      const id = request.getOdataRequestId()
      const response = responsesInBatch.get(id)
      let atomicityGroup = response.getAtomicityGroupId()

      if (atomicityGroup) {
        if (atomicityGroup !== lastUsedAtomicityGroup) {
          lastUsedAtomicityGroup = atomicityGroup
          grp = []
          list.push(grp)
        }
        grp.push(response)
      } else {
        list.push(response)
      }
      // If continue-on-error is not set, exit after the first failed request.
      if (!this._batchContext.isContinueOnError() && this._batchContext.isRequestFailed(id)) break
    }

    return list
  }

  /**
   * Writes one batched response as new part into the multipart document.
   * @param {MultipartWriter} multipartWriter Writer of the multipart document
   * @param {OdataResponseInBatch} response Response to be written
   * @param {boolean} writeContentId ContentId should be written only in changesets
   * @private
   */
  _writeResponse (multipartWriter, response, writeContentId) {
    multipartWriter.startPart()
    const applicationHttpWriter = new ApplicationHttpWriter(this._outgoingResponse)
    const contentID = response.getOdataResponseId()
    if (writeContentId && contentID) {
      applicationHttpWriter.writeHeader('content-id', contentID)
    }
    applicationHttpWriter.write(response.getIncomingResponse())
    multipartWriter.endPart()
  }

  /**
   * Writes all batched responses with help of an multipart writer into the outgoing response
   */
  execute () {
    const sendList = this._createSendList()

    // write multipart containing the the whole payload
    let multipartWriter = new MultipartWriter(this._outgoingResponse, false, this._batchContext.getBoundary())

    for (const entry of sendList) {
      if (!Array.isArray(entry)) {
        // write normal response in batch
        this._writeResponse(multipartWriter, entry)
      } else {
        // write atomicity group/change set into a nested multipart

        // get atomicity group ID
        const atomicityGroup = entry[0].getAtomicityGroupId()

        if (this._batchContext.isAtomicityGroupExecuted(atomicityGroup)) {
          // ok case
          multipartWriter.startPart()
          const atomicityGroupWriter = new MultipartWriter(this._outgoingResponse, true, atomicityGroup)
          for (const atomicityGroupEntry of entry) {
            this._writeResponse(atomicityGroupWriter, atomicityGroupEntry, true)
          }
          atomicityGroupWriter.writeEndBoundary()
          multipartWriter.endPart()
        } else {
          // atomicity group failed, the the error response replaces the change set
          const response = this._batchContext.getFailedRequestsOfAtomicityGroup(atomicityGroup)[0].getResponse()
          this._writeResponse(multipartWriter, response, true)
        }
      }
    }

    multipartWriter.writeEndBoundary(true)
  }
}

module.exports = BatchMultipartSerializer
