const { getKeyValuePair } = require('./key-value-utils')
const { validationChecks } = require('./validation-checks')
const { checkComplexType } = require('../../../util/assert')
const { deepCopyObject, deepCopyArray } = require('../../../../common/utils/copy')

const _getKeyValues = (event, target, keys) => {
  if (event === 'CREATE' && !keys) return {}
  return getKeyValuePair(target, keys)
}

const _isPrimitiveProperty = ([key], elements) => {
  return elements[key] && !elements[key].isAssociation
}

const _addPrimitiveProperty = (row, [key, value]) => {
  row[key] = value
}

const _addComplexProperty = (row, [structuredName, structuredValue], elements) => {
  // Limited to depth 1, same as for checkComplexType
  const nestedObj = {}
  for (const [structuredElement, structuredElementValue] of Object.entries(structuredValue)) {
    if (elements[`${structuredName}_${structuredElement}`]) {
      nestedObj[structuredElement] = structuredElementValue
    }
  }
  row[structuredName] = nestedObj
}

const _isCompositionProperty = ([key, value], elements) => {
  return elements[key] && elements[key].isComposition && value
}

const _addCompositionProperty = (row, [key, value], elements) => {
  const target = elements[key]._target

  if (Array.isArray(value)) {
    row[key] = []
    for (const child of value) {
      const filteredChild = {}
      _copyAndFilterDeep(filteredChild, [child], target)
      row[key].push(filteredChild)
    }
  } else {
    row[key] = {}
    _copyAndFilterDeep(row[key], [value], target)
  }
}

const _copyAndFilterDeep = (copy, data, entity) => {
  if (!Array.isArray(data)) data = [data]

  data.forEach(dataEntry => {
    let row
    if (Array.isArray(copy)) {
      row = {}
      copy.push(row)
    } else {
      row = copy
    }

    for (const prop of Object.entries(dataEntry)) {
      const elements = entity.elements

      if (_isPrimitiveProperty(prop, elements)) {
        _addPrimitiveProperty(row, prop)
      } else if (checkComplexType(prop, elements, false)) {
        _addComplexProperty(row, prop, elements)
      } else if (_isCompositionProperty(prop, elements)) {
        _addCompositionProperty(row, prop, elements)
      }
    }
  })
}

const _getFilteredCopy = (target, data, keyValues) => {
  let copy

  if (!target || target._hasPersistenceSkip) {
    // > unbound or @cds.persistence.skip
    copy = Array.isArray(data) ? deepCopyArray(data) : deepCopyObject(data)
  } else {
    // > bound (incl. CRUD) and no @cds.persistence.skip
    copy = Array.isArray(data) ? [] : {}
    _copyAndFilterDeep(copy, data, target)
  }

  if (keyValues) {
    Array.isArray(data) ? Object.assign(copy[0], keyValues) : Object.assign(copy, keyValues)
  }

  return copy
}

module.exports = (parsed, restReq) => {
  const { event, kind, segments, target, params } = parsed

  let data, keyValues
  if (kind) {
    // > action or function
    data = params || restReq.body || {}
  } else {
    // > CRUD
    keyValues = _getKeyValues(event, target, segments[1])
    if (event === 'CREATE') {
      // > batch
      data = Array.isArray(restReq.body) ? restReq.body : [restReq.body || {}]
    } else {
      data = restReq.body || {}
    }
  }

  const err = validationChecks(event, data, target && target.elements ? target : { elements: segments[0].params })

  if (err) return [err]
  else return [0, _getFilteredCopy(target, data, keyValues)]
}
