const cds = require('../../../cds')

const { SELECT } = cds.ql

const { foreignKeyPropagations } = require('../../../common/utils/foreignKeyPropagations')
const { checkReferenceIntegrity } = require('../../util/assert')
const { processDeep, processDeepAsync } = require('../../util/dataProcessUtils')
const { ensureNoDraftsSuffix, ensureDraftsSuffix, draftIsLocked } = require('../../../fiori/utils/handler')
const { isCustomOperation } = require('../../adapter/odata-v4/utils/request')

const { DRAFT_COLUMNS, DRAFT_COLUMNS_ADMIN } = require('../../../common/constants/draft')

const _isAssociationToOneManaged = element => element._isAssociationEffective && element.is2one && !element.on

const _prefixDraftColumns = () => {
  return DRAFT_COLUMNS_ADMIN.map(col => {
    return { ref: ['DRAFT_DraftAdministrativeData', col] }
  })
}

const _getSelectDraftDataCqn = (entityName, where) => {
  return SELECT.from(ensureDraftsSuffix(entityName), _prefixDraftColumns())
    .join('DRAFT.DraftAdministrativeData')
    .on('DraftAdministrativeData_DraftUUID =', { ref: ['DRAFT.DraftAdministrativeData', 'DraftUUID'] })
    .where(where)
}

const _getWheres = (key, data) => {
  const wheres = []
  for (const d of data) {
    wheres.push({ [key.name]: d[key.name] })
  }
  return wheres
}

const allKeysAreProvided = req => {
  const data = req.data && (Array.isArray(req.data) ? req.data : [req.data])
  for (const key of Object.values(req.target.keys)) {
    if (key._isAssociationStrict || DRAFT_COLUMNS.includes(key.name)) {
      continue
    }
    for (const d of data) {
      if (d[key.name] === undefined) return false
    }
  }
  return true
}

const _addWhereToCqn = (wheres, target, cqn) => {
  if (wheres.length === 1) {
    const whereObj = wheres[0]
    for (const key in whereObj) {
      // check if we have structured elements
      if (target.elements[key].elements) {
        whereObj[key] = { val: JSON.stringify(whereObj[key]) }
      }
    }
    cqn.where(whereObj)
  } else {
    cqn.where({ or: wheres })
  }
}

const _getSelectCQN = (req, columns) => {
  const cqn = SELECT.from(req.target)

  if (columns) {
    cqn.columns(...columns)
  }

  const data = req.data && (Array.isArray(req.data) ? req.data : [req.data])

  for (const key of Object.values(req.target.keys)) {
    if (key._isAssociationStrict || DRAFT_COLUMNS.includes(key.name)) {
      continue
    }

    const wheres = _getWheres(key, data)
    if (wheres.length) {
      _addWhereToCqn(wheres, req.target, cqn)
    }
  }

  if (req.target.query && req.target.query.SELECT && req.target.query.SELECT.orderBy) {
    cqn.SELECT.orderBy = req.target.query.SELECT.orderBy
  }

  return cqn
}

const validateDraft = (draftResult, req) => {
  if (!draftResult || draftResult.length === 0) {
    req.reject(404)
  }

  const draftAdminData = draftResult[0]

  // the same user that locked the entity can always delete it
  if (draftAdminData.InProcessByUser === req.user.id) {
    return
  }

  // proceed with the delete action only if it was initiated by a different user
  // than the one who locked the entity and the configured drafts cancellation
  // timeout timer has expired
  if (draftIsLocked(draftAdminData.LastChangeDateTime)) {
    req.reject(403, 'DRAFT_LOCKED_BY_ANOTHER_USER')
  }
}

function _fillForeignKeysWithNull(managedAssocToOneElement, row) {
  const keys = foreignKeyPropagations(managedAssocToOneElement)
  for (const key of keys) {
    row[key.parentFieldName] = null
  }
}

const _flattenManagedToOneAssociation = (managedAssocToOneElement, entity, row, property, csn, req) => {
  const targetEntity = managedAssocToOneElement._target
  for (const key in targetEntity.keys) {
    const el = targetEntity.keys[key]
    if (_isAssociationToOneManaged(el)) {
      if (row[managedAssocToOneElement.name] === null) {
        _fillForeignKeysWithNull(managedAssocToOneElement, row)
        continue
      }
      _flattenManagedToOneAssociation(el, targetEntity, row[managedAssocToOneElement.name], key, csn, req)
      continue
    }

    if (row[managedAssocToOneElement.name] === null) {
      row[managedAssocToOneElement.name + '_' + key] = null
    } else {
      row[managedAssocToOneElement.name + '_' + key] = row[managedAssocToOneElement.name][key]
      delete row[managedAssocToOneElement.name][key]
    }
  }

  if (row[managedAssocToOneElement.name] === null || Object.keys(row[managedAssocToOneElement.name]).length === 0) {
    // if there are no non key values left, remove assoc object
    delete row[managedAssocToOneElement.name]
  }
}

const _flattenDeepToOneAssociations = (entity, data, csn, req) => {
  if (!Array.isArray(data)) {
    return _flattenDeepToOneAssociations(entity, [data], csn, req)
  }

  for (const row of data) {
    for (const property in row) {
      const element = entity.elements[property]
      if (!element) continue

      if (_isAssociationToOneManaged(element)) {
        _flattenManagedToOneAssociation(element, entity, row, property, csn, req)
      } else if (element.elements) {
        _flattenDeepToOneAssociations(element, [row[element.name]], csn, req)
      }
    }
  }
}

const flattenDeepToOneAssociations = (req, csn) => {
  if (!req.target) {
    return
  }

  if (req.event !== 'CREATE' && req.event !== 'UPDATE') {
    return
  }

  // REVISIT: adopt template mechanism?
  processDeep(
    (data, entity) => {
      _flattenDeepToOneAssociations(entity, data, csn, req)
    },
    req.data,
    req.target,
    false,
    true
  )
}

const checkIntegrityWrapper = (req, csn, run) => async (data, entity) => {
  const errors = await checkReferenceIntegrity(entity, data, req, csn, run)
  if (errors && errors.length !== 0) {
    for (const err of errors) {
      req.error(err)
    }
  }
}

const _isUncheckableInsert = query => {
  return query.INSERT && (query.INSERT.rows || query.INSERT.values || query.INSERT.as)
}

// REVISIT: lower to db layer, where it's used
const checkIntegrityUtil = async (req, csn, run) => {
  if (!run) return

  // REVISIT
  if (typeof req.query === 'string' || req.target._unresolved) return

  // FIXME: doesn't work for uncheckable inserts
  if (_isUncheckableInsert(req.query)) return

  // REVISIT: integrity check needs context.data
  if (Object.keys(req.data).length === 0) {
    // REVISIT: We may need to double-check re req.data being undefined or empty
    if (req.query.DELETE) {
      req.data = req._beforeDeleteData || {}
    } else if (req.context && req.context.data && Object.keys(req.context.data).length > 0) {
      req.data = req.context.data
    }
  }
  if (Object.keys(req.data).length === 0) return

  await processDeepAsync(checkIntegrityWrapper(req, csn, run), req.data, req.target, false, true)
}

const _addDraftDataToContext = (req, result) => {
  validateDraft(result, req)

  if (req.rejected) {
    return
  }

  if (!req._draftMetadata) {
    req._draftMetadata = {}
  }

  DRAFT_COLUMNS_ADMIN.forEach(column => {
    if (column in result[0]) req._draftMetadata[column] = result[0][column]
  })

  req.data.DraftAdministrativeData_DraftUUID = result[0].DraftUUID
}

// copied from adapter/odata-v4/utils/context-object
const _getTargetEntityName = (service, pathSegments) => {
  if (isCustomOperation(pathSegments, false)) {
    return undefined
  }

  let navSegmentName
  let entityName = `${service.name}.${pathSegments[0].getEntitySet().getName()}`

  for (const navSegment of pathSegments.filter(segment => segment.getNavigationProperty() !== null)) {
    navSegmentName = navSegment.getNavigationProperty().getName()
    entityName = service.model.definitions[entityName].elements[navSegmentName].target
  }

  return entityName
}

/**
 * Provide information about the parent entity, i.e. the entity that has the to-many composition element.
 * Limitation: only works for one key (besides IsActiveEntity)
 *
 * @param service
 * @param req
 * @returns {object}
 * @private
 */
const _getParent = (service, req) => {
  // REVISIT: get rid of getUriInfo
  if (!req.getUriInfo) return

  const segments = req.getUriInfo().getPathSegments()

  if (segments.length === 1) return

  const parent = {
    entityName: _getTargetEntityName(service, segments.slice(0, segments.length - 1))
  }

  const parentKeyPredicates = segments[segments.length - 2].getKeyPredicates()
  let keyPredicateName, keyPredicateText
  for (const keyPredicate of parentKeyPredicates) {
    keyPredicateName = keyPredicate.getEdmRef().getName()
    keyPredicateText = keyPredicate.getText()

    if (keyPredicateName === 'IsActiveEntity') {
      parent.IsActiveEntity = keyPredicateText
    } else {
      parent.keyName = keyPredicateName
      parent.keyValue = keyPredicateText
    }
  }

  return parent
}

const addDraftDataFromExistingDraft = async (req, service) => {
  const parent = _getParent(service, req)
  let result

  if (parent && parent.IsActiveEntity === 'false') {
    const parentWhere = [{ ref: [parent.keyName] }, '=', { val: parent.keyValue }]
    result = await cds.tx(req).run(_getSelectDraftDataCqn(parent.entityName, parentWhere))
    _addDraftDataToContext(req, result)
    return result
  }

  if (!parent) {
    const keys = Object.keys(req.target.keys)
    const rootWhere = keys.reduce((res, key) => {
      if (key === 'IsActiveEntity') {
        return res
      }
      res[key] = req.data[key]
      return res
    }, {})

    result = await cds.tx(req).run(_getSelectDraftDataCqn(ensureNoDraftsSuffix(req.target.name), rootWhere))
    if (result && result.length > 0) {
      _addDraftDataToContext(req, result)
    }
    return result
  }

  return []
}

const addGeneratedDraftUUID = async req => {
  req._draftMetadata = req._draftMetadata || {}
  req.data.DraftAdministrativeData_DraftUUID = cds.utils.uuid()
  req._draftMetadata.DraftUUID = req.data.DraftAdministrativeData_DraftUUID
}

/*
 * merge CQNs
 */
const _mergeExpandCQNs = cqns => {
  const cols = cqns[0].SELECT.columns
  for (const c of cqns) {
    for (const col of c.SELECT.columns) {
      if (col.expand && !cols.find(ele => ele.ref[0] === col.ref[0])) {
        cols.push(col)
      }
    }
  }
  return cqns[0]
}

/*
 * recursively builds a select cqn for deep read after write
 * (depth determined by req.data)
 */
const getDeepSelect = req => {
  // REVISIT: Why do we do such expensive deep reads after write at all ???
  const { elements } = req.target
  const cols = []

  for (const each in elements) {
    if (DRAFT_COLUMNS.includes(each)) continue
    const e = elements[each]
    if (!e.isAssociation) cols.push(each)
    else if (e._isCompositionEffective) {
      const cqns = []
      // REVISIT: This recursively creates lots of nested selects only to keep a few
      for (const row of Array.isArray(req.data) ? req.data : [req.data]) {
        let d = row[each]
        if (!d) continue
        else if (!Array.isArray(d)) d = [d]
        else if (d.length === 0) continue
        const nested = d.map(data => getDeepSelect({ target: e._target, data }))
        cqns.push(_mergeExpandCQNs(nested))
      }
      if (cqns.length > 0) {
        const { SELECT } = _mergeExpandCQNs(cqns)
        cols.push({ ref: [each], expand: SELECT.columns })
      }
    }
  }

  // root? -> with where clause
  return req.event ? _getSelectCQN(req, cols) : SELECT.from(req.target, cols)
}

module.exports = {
  addDraftDataFromExistingDraft,
  addGeneratedDraftUUID,
  getDeepSelect,
  allKeysAreProvided,
  checkIntegrityUtil,
  flattenDeepToOneAssociations
}
