const Where = require('./Where')
const { invalidFunctionArgumentError } = require('../util/errors')

const fnChain = Symbol.for('sap.cds.fnChain')

/**
 * UPDATE statement updates values or rows from a specific table.
 *
 * @extends {Where}
 */
class Update extends Where {
  constructor() {
    super('UPDATE')
  }

  static entity(entity) {
    const cqn = new Update()
    cqn[fnChain] = cqn[fnChain].concat('()') // UPDATE.entity() is not exposed but UPDATE()

    if (typeof entity === 'string' && (entity.includes(':') || entity.includes('['))) {
      cqn.UPDATE.entity = Where.convertPathExpression(entity)
    } else {
      cqn.UPDATE.entity = Where._isEntity(entity, 'UPDATE()')
    }

    return cqn
  }

  _distribute(obj) {
    this.UPDATE.data = this.UPDATE.data || {}
    this.UPDATE.with = this.UPDATE.with || {}

    for (const k in obj) {
      if ('val' in obj[k]) {
        this.UPDATE.data[k] = obj[k].val
      } else {
        this.UPDATE.with[k] = obj[k]
      }
    }

    return this
  }

  /**
   * Sets the values that should be updated
   *
   * @example
   * UPDATE('Authors').set({NAME: 'Jon Doe', STOCK: 123, count: {func: 'count', args: [{ ref: ['abc'] }], as: 'count'}})
   * @param {object} valueObject contains the properties that should be updated, can also contain expressions
   * @returns {object} UPDATE statement
   * @throws Error - If no or invalid input parameter provided
   */
  set(valueObject) {
    this[fnChain] = this[fnChain].concat('.set()')

    const checkIsInvalid = this._isInvalidValueObject(valueObject, true)
    if (checkIsInvalid.status) {
      throw invalidFunctionArgumentError(this[fnChain], checkIsInvalid.data)
    }

    return this._distribute(this._convertToCqnObject(valueObject))
  }

  /**
   * Sets the values that should be updated, no expressions allowed
   *
   * @example
   * UPDATE('Authors').data({NAME: 'Jon Doe', STOCK: 123})
   * @param {object} valueObject contains the properties that should be updated
   * @returns {object} UPDATE statement
   * @throws Error - If no or invalid input parameter provided
   */
  data(valueObject) {
    this[fnChain] = this[fnChain].concat('.data()')

    const checkIsInvalid = this._isInvalidValueObject(valueObject)
    if (checkIsInvalid.status) {
      throw invalidFunctionArgumentError(this[fnChain], checkIsInvalid.data)
    }

    this.UPDATE.data = valueObject
    return this
  }

  /**
   * Sets the values that should be updated
   *
   * @example
   * UPDATE('Authors').set({NAME: 'Jon Doe', STOCK: 123, count: {func: 'count', args: [{ ref: ['abc'] }], as: 'count'}})
   *
   * @param {object} valueObject contains the properties that should be updated, can also contain expressions
   */
  with(valueObject) {
    this[fnChain] = this[fnChain].concat('.with()')
    return this.set(valueObject)
  }

  _convertToCqnObject(valueObject) {
    const cqnObj = {}
    for (const key in valueObject) {
      cqnObj[key] = this._generateValObj(valueObject[key], key)
    }
    return cqnObj
  }

  _generateValObj(value, key) {
    if (value === null || value === undefined) {
      return { val: value }
    }

    if (value instanceof Buffer) {
      return { val: value }
    }

    if (['-=', '+=', '*=', '/=', '%='].includes(Object.keys(value)[0])) {
      const op = Object.keys(value)[0]
      const val = value[op]
      return { xpr: [{ ref: [key] }, op.substring(0, 1), { val: val }] }
    }

    if (value['=']) {
      return { ref: [value['=']] }
    }

    // value.pipe is a stream indicator
    if (!Array.isArray(value) && typeof value === 'object' && typeof value.pipe !== 'function') {
      const cqnObj = {}
      for (const key in value) {
        cqnObj[key] = this._generateValObj(value[key], key)
      }
      return cqnObj
    }

    return { val: value }
  }

  _isEmptyObject(value) {
    return Object.keys(value).length === 0
  }

  _isInvalidValueObject(valueObject) {
    if (!valueObject || typeof valueObject !== 'object' || Array.isArray(valueObject)) {
      return { status: true, data: valueObject }
    }

    const invalidEntries = {}

    for (const key in valueObject) {
      if (
        valueObject[key] === undefined ||
        (valueObject[key] &&
          typeof valueObject[key] === 'object' &&
          !Buffer.isBuffer(valueObject[key]) &&
          !Array.isArray(valueObject[key]) &&
          !(valueObject[key] instanceof Date) &&
          this._isEmptyObject(valueObject[key]))
      ) {
        invalidEntries[key] = valueObject[key]
      }
    }

    if (Object.keys(invalidEntries).length !== 0) {
      return { status: true, data: invalidEntries }
    }

    return { status: false }
  }

  valueOf() {
    return `UPDATE ${Update._quoteElement(this.UPDATE.entity)} `
  }
}

module.exports = Update
