const cds = require('../../cds')
const LOG = cds.log('app')

const _require = require('../utils/require')
const { UNAUTHORIZED } = require('../utils/auth')

let passport

// strategy initializers for lazy loading of dependencies
const _initializers = {
  // REVISIT: support basic authentication?
  basic: ({ credentials }) => {
    const BasicStrategy = require('./strategies/basic')
    passport.use(new BasicStrategy(credentials))
  },
  dummy: () => {
    const DummyStrategy = require('./strategies/dummy')
    passport.use(new DummyStrategy())
  },
  JWT: ({ uaa }) => {
    const JWTStrategy = require('./strategies/JWT')
    passport.use(new JWTStrategy(uaa))
  },
  mock: ({ users }, srvName) => {
    const MockStrategy = require('./strategies/mock')
    passport.use(new MockStrategy(users, `mock_${srvName}`))
  },
  xsuaa: ({ uaa }) => {
    const XSUAAStrategy = require('./strategies/xsuaa')
    passport.use(new XSUAAStrategy(uaa))
  }
}

// map for initialized authenticators
const _authenticators = {}

const _isRestricted = srv => {
  return !!(
    srv.definition['@requires'] ||
    Object.keys(srv.entities).some(k => srv.entities[k]['@requires'] || srv.entities[k]['@restrict']) ||
    Object.keys(srv.entities).some(
      k =>
        srv.entities[k].actions &&
        Object.keys(srv.entities[k].actions).some(
          l => srv.entities[k].actions[l]['@requires'] || srv.entities[k].actions[l]['@restrict']
        )
    ) ||
    Object.keys(srv.operations).some(k => srv.operations[k]['@requires'] || srv.operations[k]['@restrict'])
  )
}

const _initializeStrategy = (strategy, config, srv) => {
  if (!_initializers[strategy]) {
    // REVISIT: why?
    process.exitCode = 1
    throw new Error(`Authentication strategy "${strategy}" is not supported`)
  }

  if (!_authenticators[strategy] || strategy === 'mock' || process.env.NODE_ENV === 'test') {
    _initializers[strategy](config, srv.name)
    _authenticators[strategy] = true
  }
}

const _callback = (req, res, next, err, user, info) => {
  // REVISIT: when is this the case and what to do?
  if (err) return res.status(401).json(UNAUTHORIZED) // > no details to client

  let challenges
  if (info && Array.isArray(info)) {
    // > info === challenges
    challenges = info.filter(ele => ele)
    challenges = challenges.length ? challenges : undefined
    info = null
  }

  // compat req._.req.authInfo
  if (info) req.authInfo = info

  req.user = user || Object.defineProperty(new cds.User(), '_challenges', { enumerable: false, value: challenges })
  Object.defineProperty(req.user, '_req', { enumerable: false, value: req })

  next()
}

module.exports = (srv, app, options) => {
  let config = options.auth

  const isRestricted = _isRestricted(srv)
  const isMultiTenant = !!(cds.env.requires && cds.env.requires.db && cds.env.requires.db.multiTenant)

  if (!config && !isRestricted && (!isMultiTenant || process.env.NODE_ENV !== 'production')) {
    if (isMultiTenant) {
      LOG._warn && LOG.warn(`[${srv.name}] - Authentication needed for multitenancy in production.`)
    }

    return
  }

  config = config || cds.env.requires.auth

  if (config.impl) {
    // > custom middleware
    app.use(srv.path, _require(cds.resolve(config.impl)))

    return
  }

  // here, we need passport
  passport = passport || _require('passport')

  // initialize strategies
  config.strategy = Array.isArray(config.strategy) ? config.strategy : [config.strategy]
  for (const strategy of config.strategy) {
    _initializeStrategy(strategy, config, srv)
  }

  // authenticate
  app.use(srv.path, passport.initialize())
  app.use(srv.path, (req, res, next) => {
    passport.authenticate(
      config.strategy.map(s => (s === 'mock' ? `mock_${srv.name}` : s)),
      { session: false, failWithError: true },
      _callback.bind(undefined, req, res, next)
    )(req, res, next)
  })
}
