const BaseBuilder = require('./BaseBuilder')

/**
 * DeleteBuilder is used to take a CQN object as an input and to build an object representing a delete operation
 * with SQL string and values to be used with a prepared statement.
 * The SQL object can only be built if the property 'from' is available.
 * The property 'where' is optional.
 *
 * @example <caption>Example of CQN </caption>
 * {
 *  DELETE = {DELETE:{
 *  from: entity | string,
 *  where: _xpr
 *  }}
 * }
 */
class DeleteBuilder extends BaseBuilder {
  get ExpressionBuilder() {
    const ExpressionBuilder = require('./ExpressionBuilder')
    Object.defineProperty(this, 'ExpressionBuilder', { value: ExpressionBuilder })
    return ExpressionBuilder
  }

  get ReferenceBuilder() {
    const ReferenceBuilder = require('./ReferenceBuilder')
    Object.defineProperty(this, 'ReferenceBuilder', { value: ReferenceBuilder })
    return ReferenceBuilder
  }

  /**
   * Builds an Object based on the properties of the CQN object.
   *
   * @example <caption>Example output</caption>
   * {
   *    sql: 'DELETE FROM "T" WHERE "COLUMN" = ?',
   *    values: [1]
   * }
   *
   * @returns {{sql: string, values: Array}} Object with two properties.
   * SQL string for prepared statement and array of values to replace the placeholders.
   */
  build() {
    this._outputObj = {
      sql: ['DELETE'],
      values: []
    }
    this._from()
    if (Array.isArray(this._obj.DELETE.where) && this._obj.DELETE.where.length > 0) {
      this._where()
    }

    this._outputObj.sql = /** @type {string} */ this._outputObj.sql.join(' ')
    return this._outputObj
  }

  _from() {
    this._outputObj.sql.push('FROM')
    if (this._obj.DELETE.from.ref) {
      const dbName = this._getDatabaseName(this._obj.DELETE.from.ref[0])
      // TODO: was this ever needed?
      // const res = new this.ReferenceBuilder(this._obj.DELETE.from, this._options, this._csn).build()
      this._outputObj.sql.push(this._quoteElement(dbName))
      this._options.entityName = this._obj.DELETE.from.ref[0]

      if (this._obj.DELETE.from.as) {
        // identifier
        this._outputObj.sql.push('as', this._quoteElement(this._obj.DELETE.from.as))
      }
    } else if (typeof this._obj.DELETE.from === 'string') {
      const dbName = this._getDatabaseName(this._obj.DELETE.from)
      this._outputObj.sql.push(this._quoteElement(dbName))
      this._options.entityName = this._obj.DELETE.from
    } else {
      const dbName = this._getDatabaseName(this._obj.DELETE.from.name)
      this._outputObj.sql.push(this._quoteElement(dbName))
      this._options.entityName = this._obj.DELETE.from.name
    }
  }

  _where() {
    const where = new this.ExpressionBuilder(this._obj.DELETE.where, this._options, this._csn).build()
    this._outputObj.sql.push('WHERE', where.sql)
    this._outputObj.values = where.values
  }
}

module.exports = DeleteBuilder
