const BaseBuilder = require('./BaseBuilder')

/**
 * DropBuilder is used to build an object, representing a drop operation
 *
 * @example <caption>Example of CQN </caption>
 * {
 *  DROP = {DROP:{
 *  table: entity | string,
 *  view: entity | string,
 *  entity: entity | string
 *  }}
 * }
 */
class DropBuilder extends BaseBuilder {
  /**
   * Builds an Object based on the properties of the CQN object.
   *
   * @example <caption>Example output</caption>
   * {
   *    sql: 'DROP TABLE IF EXISTS name'
   * }
   *
   * @returns {{sql: string}} Object containing sql.
   */
  build() {
    this._sql = ['DROP']
    this._tableOrView()
    this._ifExists()
    this._getName()

    return {
      sql: this._sql.join(' ')
    }
  }

  /**
   * Can be overridden to add custom logic in case DB does not support exists
   *
   * @private
   */
  _ifExists() {
    this._sql.push('IF EXISTS')
  }

  _tableOrView() {
    let type = 'TABLE'
    if (this._obj.DROP.view) {
      type = 'VIEW'
    } else if (!this._obj.DROP.table && typeof this._obj.DROP.entity === 'object') {
      type = this._obj.DROP.entity.query || this._obj.DROP.entity.type ? 'VIEW' : 'TABLE'
    }

    this._sql.push(type)
  }

  _getName() {
    const entity = this._obj.DROP.view || this._obj.DROP.table || this._obj.DROP.entity
    this._sql.push(this._quoteElement(typeof entity === 'string' ? entity : entity.name))
  }
}

module.exports = DropBuilder
