const BaseBuilder = require('./BaseBuilder')

/**
 * ReferenceBuilder is used to take a part of a CQN object as an input and to build an object representing a reference
 * with SQL string and values.
 *
 * Currently it supports the references like below:
 *
 * @example <caption>Simple ref part of CQN </caption>
 * {ref: ['x']}
 * {ref: ['func_name', { args: [func_args] }]}
 */
class ReferenceBuilder extends BaseBuilder {
  get FunctionBuilder() {
    const FunctionBuilder = require('./FunctionBuilder')
    Object.defineProperty(this, 'FunctionBuilder', { value: FunctionBuilder })
    return FunctionBuilder
  }

  /**
   * Builds an Object based on the properties of the input object in the constructor.
   *
   * @example <caption>Example output</caption>
   * {
   *    sql: '"X"',
   *    values: []
   * }
   * {
   *    sql: '"func_name(?,?)"',
   *    values: [1, 'a']
   * }
   *
   * @returns {{sql: string, values: Array}} Object with two properties.
   * SQL string for prepared statement and an empty array of values.
   */
  build() {
    this._outputObj = {
      sql: [],
      values: []
    }

    if (this._isFunction()) {
      const { sql, values } = new this.FunctionBuilder(this._obj, this._options, this._csn).build()
      this._outputObj.sql.push(sql)
      this._outputObj.values.push(...values)
    } else if (this._obj.ref) {
      // reference
      if (this._obj.param) {
        this._parseParamReference(this._obj.ref)
      } else {
        this._parseReference(this._obj.ref)
      }
    } else {
      this._outputObj.sql.push(this._obj)
    }

    this._sortOrder()
    this._outputObj.sql = this._outputObj.sql.join(' ')
    return this._outputObj
  }

  _sortOrder() {
    if (!Object.prototype.hasOwnProperty.call(this._obj, 'sort')) {
      return
    }

    if (this._options.collate) {
      this._outputObj.sql.push(this._options.collate)
    }

    this._outputObj.sql.push(this._obj.sort === 'desc' ? 'DESC' : 'ASC')
  }

  _isFunction() {
    return (this._obj.ref && this._obj.ref.length > 1 && this._obj.ref[1].args) || (this._obj.func && this._obj.args)
  }

  _parseReference(refArray) {
    if (refArray[0].id) {
      throw new Error(`${refArray[0].id}: Views with parameters supported only on HANA`)
    }

    const entity = this._csn && this._csn.definitions[this._options.entityName]
    const element = entity && entity.elements[refArray[0]]
    if (element) {
      if (element.elements || (element.is2one && !element.on)) {
        // REVISIT we assume that structured elements are already unfolded here
        this._outputObj.sql.push(refArray.join('_'))
        return
      }
    }

    this._outputObj.sql.push(refArray.map(el => this._quoteElement(el)).join('.'))
  }

  _parseParamReference(refArray) {
    if (refArray[0] === '?') {
      this._outputObj.sql.push(this._options.placeholder)
    } else {
      this._outputObj.sql.push(this._options.placeholder)
      this._outputObj.values.push(refArray[0])
    }
  }
}

module.exports = ReferenceBuilder
